from sklearn.preprocessing import MultiLabelBinarizer
import numpy as np
import pandas as pd
import pickle
from tqdm import tqdm

tqdm.pandas()


class DataframePreprocessing:

    DEFAULT_TARGET_THEMES = [
        5,
        6,
        26,
        33,
        139,
        163,
        232,
        313,
        339,
        350,
        406,
        409,
        555,
        589,
        597,
        634,
        660,
        695,
        729,
        766,
        773,
        793,
        800,
        810,
        852,
        895,
        951,
        975,
    ]

    OTHER_THEMES_VALUE = 4242

    def __init__(
        self,
        df=pd.DataFrame(),
        group_processes=True,
        x_column_name="page_text_extract",
        y_column_name="tema",
        target_themes=DEFAULT_TARGET_THEMES,
        other_themes_value=OTHER_THEMES_VALUE,
        is_incremental_training=False,
        remove_processes_without_theme=True,
        is_parquet=False,
        labels_freq={},
        vocab_path="",
    ):
        self.is_incremental_training = is_incremental_training
        self.remove_processes_without_theme = remove_processes_without_theme
        self.is_parquet = is_parquet
        self.x_column_name = x_column_name
        self.y_column_name = y_column_name
        self.other_themes_value = other_themes_value
        self.target_themes = target_themes
        self.group_processes = group_processes
        self.vocab_path = vocab_path

        self.distinct_themes = target_themes + [other_themes_value]
        if not df.empty:

            if remove_processes_without_theme:
                df = df[
                    df[self.y_column_name].progress_apply(
                        self._remove_processes_without_theme
                    )
                ].copy()
                df[self.y_column_name] = (
                    df[self.y_column_name]
                    .progress_apply(self._remove_with_without_theme_mixture)
                    .copy()
                )

            self._generate_themes_column(df.copy())

            self.target_themes.sort()
            self.labels_freq = {}

            if not labels_freq:
                self._set_labels_frequency()
            else:
                self.labels_freq = labels_freq
            self.processed_df = self._process_dataframe()

    def _remove_processes_without_theme(self, series):
        if 0 in series and len(series) == 1:
            return False
        return True

    def _remove_with_without_theme_mixture(self, series):
        if 0 in series and len(series) > 1:
            return np.array([v for v in series if v != 0])
        return series

    def _generate_themes_column(self, df):
        if self.group_processes:
            self._group_samples_by_process(df)
        else:
            if not self.is_parquet:
                self.df = df
                # self.df[self.y_column_name] = self.df[self.y_column_name].apply(lambda x: np.array(x))
                # self._transform_array_column(df)
            else:
                self.df = df
                self.df[self.y_column_name] = self.df[self.y_column_name].apply(
                    lambda x: np.array(x)
                )

    def _transform_array_column(self, df):
        print("Tranforming themes strings to arrays...")
        df[self.y_column_name] = df[self.y_column_name].apply(
            lambda l: np.fromstring(l[1:-1], sep=" ")
        )
        self.df = df

    def _group_samples_by_process(self, df):
        print("Grouping processes...")
        self.df = df.groupby("process_id").apply(self._aggregate)

    def _aggregate(self, series):
        reduced = {}
        series[self.x_column_name].drop_duplicates(inplace=True)

        reduced[self.x_column_name] = " ".join(
            str(x) for x in series[self.x_column_name].values
        )
        temas = np.unique(series[self.y_column_name].values)
        reduced[self.y_column_name] = temas[~np.isnan(temas)]
        return pd.Series(reduced, index=[self.x_column_name, *[self.y_column_name]])

    def _remove_rare_samples(self, series, threshold=5):
        if self.labels_freq.get(tuple(series.tolist())) < threshold:
            return False
        return True

    def _switch_other_themes_values(self, actual_label):
        """
            Replace the values of themes that are not in target themes
            to a unique value
        """

        modified_label = set()
        for theme in actual_label:
            if theme not in self.target_themes:
                modified_label.add(self.other_themes_value)
            elif theme != 0:
                modified_label.add(theme)
        return sorted(modified_label)

    def _normalize_labels(self, series):
        return np.asarray(self._switch_other_themes_values(series.tolist()))

    def _set_labels_frequency(self):
        print("Setting labels frequency...")
        for label in self.df[self.y_column_name]:
            normalized_label = tuple(self._switch_other_themes_values(label))

            if not self.labels_freq.get(normalized_label):
                self.labels_freq[normalized_label] = 1
            else:
                self.labels_freq[normalized_label] += 1

    def get_labels_frequency(self, series):
        print("Setting labels frequency...")
        labels_freq = {}
        for label in series:
            normalized_label = tuple(self._switch_other_themes_values(label))

            if not labels_freq.get(normalized_label):
                labels_freq[normalized_label] = 1
            else:
                labels_freq[normalized_label] += 1
        return labels_freq

    def _get_distinct_themes(self):
        if self.is_incremental_training:
            self.distinct_themes = self.target_themes + [self.other_themes_value]
        else:
            distinct_themes = set()
            for label in self.df["labels_with_others"]:
                for theme in label:
                    distinct_themes.add(theme)
            self.distinct_themes = list(sorted(distinct_themes))

    # TODO: Include case when themes are not grouped
    def get_unique_binarized_labels(self, df_path, y_column_name, is_parquet=False):
        print("Generating set of binarized labels...")
        if not is_parquet:
            themes = pd.read_csv(df_path, usecols=[y_column_name])
            themes[y_column_name] = themes[y_column_name].apply(
                lambda l: np.fromstring(l[1:-1], sep=" ")
            )
        else:
            themes = pd.read_parquet(df_path, columns=[y_column_name])
            themes[self.y_column_name] = themes[self.y_column_name].apply(
                lambda x: np.array(x)
            )
            themes[self.y_column_name] = themes[self.y_column_name].apply(
                self._remove_with_without_theme_mixture
            )

        if self.remove_processes_without_theme:
            themes = themes[
                themes[self.y_column_name].progress_apply(
                    self._remove_processes_without_theme
                )
            ]

        labels_freq = self.get_labels_frequency(themes[self.y_column_name])

        themes["labels_with_others"] = themes[y_column_name].apply(
            self._normalize_labels
        )

        mlb = MultiLabelBinarizer()
        binarized_columns = mlb.fit_transform(themes["labels_with_others"].to_numpy())
        unique_labels = []

        for bin_label in binarized_columns:
            if bin_label.tolist() not in unique_labels:
                unique_labels.append(bin_label.tolist())
        return unique_labels, labels_freq

    def _binarize_labels(self):
        print("Binarizing labels...")
        self._get_distinct_themes()

        mlb = MultiLabelBinarizer(classes=self.distinct_themes)
        binarized_columns = mlb.fit_transform(self.df["labels_with_others"].to_numpy())

        columns_names = {
            ix: binarized_columns[:, i] for i, ix in enumerate(self.distinct_themes)
        }

        return pd.concat(
            [self.df.reset_index(drop=True), pd.DataFrame(columns_names)], axis=1,
        )

    def _clean_text(self, text, vocab):
        text = text.split()
        text = [x for x in text if x in vocab]
        text = " ".join(text)
        return text

    def _select_vocab(self, vocab_path):
        vocab = pickle.load(open(vocab_path, "rb"))
        self.df[self.x_column_name] = self.df[self.x_column_name].progress_apply(
            self._clean_text, vocab=vocab
        )

    def _process_dataframe(self):
        self.df = self.df[~pd.isnull(self.df[self.x_column_name])]
        self.df[self.y_column_name] = self.df[self.y_column_name].apply(
            self._remove_with_without_theme_mixture
        )

        if self.remove_processes_without_theme:
            self.df = self.df[
                self.df[self.y_column_name].progress_apply(
                    self._remove_processes_without_theme
                )
            ]

        if self.vocab_path:
            self._select_vocab(self.vocab_path)

        self.df["labels_with_others"] = self.df[self.y_column_name].apply(
            self._normalize_labels
        )
        print("Removing rare samples...")

        self.df = self.df[
            self.df["labels_with_others"].apply(self._remove_rare_samples)
        ]
        return self._binarize_labels()



"""
Module to facilitate the integration of a sklearn training pipeline into a deploy and retraining system
"""

from .multilabel_training import MultilabelTraining
from .metrics import *
from .cnn_pecas_model import PecasModel

__version__ = "0.0.10"



import numpy as np
from sklearn.metrics import classification_report, accuracy_score, \
                            f1_score, precision_score, recall_score, hamming_loss, \
                            jaccard_score


def hit_one(y_true, y_pred):
    score = []
    for pred, true in zip(y_pred, y_true):
        indexes = np.where(true == 1)
        if (pred[indexes] == true[indexes]).any():
            score.append(1)
        else:
            score.append(0)
    return sum(score) / len(score)

def mistake_one(y_true, y_pred):
    score = []
    for pred, true in zip(y_pred, y_true):
        indexes = np.where(true == 0)
        if (pred[indexes] != true[indexes]).any():
            score.append(1)
        else:
            score.append(0)
    return sum(score) / len(score)

def true_positive(y_true, y_pred):
    score = []
    for pred, true in zip(y_pred, y_true):
        indexes = np.where(true == 1)
        correct = sum(pred[indexes] == true[indexes].any())
        number_of_positives = len(true[indexes])
        if number_of_positives:
            score.append(correct / len(true[indexes]))
    return sum(score)/ len(score)

def true_negative(y_true, y_pred):
    score = []
    for pred, true in zip(y_pred, y_true):
        indexes = np.where(true == 0)
        correct = sum(pred[indexes] == true[indexes].any())
        number_of_positives = len(true[indexes])
        if number_of_positives:
            score.append(correct / len(true[indexes]))
    return sum(score)/ len(score)

def false_positive(y_true, y_pred):
    score = []
    for pred, true in zip(y_pred, y_true):
        indexes = np.where(true == 0)
        wrong = sum(pred[indexes] != true[indexes].any())
        number_of_negatives = len(true[indexes])
        if number_of_negatives:
            score.append(wrong / number_of_negatives)
        else:
            score.append(0)
    return sum(score) / len(score)

def false_negative(y_true, y_pred):
    score = []
    for pred, true in zip(y_pred, y_true):
        indexes = np.where(true == 1)
        wrong = sum(pred[indexes] != true[indexes].any())
        number_of_negatives = len(true[indexes])
        if number_of_negatives:
            score.append(wrong / number_of_negatives)
        else:
            score.append(0)
    return sum(score) / len(score)

def get_multilabel_metrics(y_test, y_pred_pac):
    one_hit = hit_one(y_test.values,y_pred_pac)
    one_mistake = mistake_one(y_test.values, y_pred_pac)
    tp = true_positive(y_test.values, y_pred_pac)
    fp = false_positive(y_test.values, y_pred_pac)
    tn = true_negative(y_test.values, y_pred_pac)
    fn = false_negative(y_test.values, y_pred_pac)
    hm = hamming_loss(y_test, y_pred_pac)
    print('Métricas para Passive Agressive Classifier:\n')
    print("Pelo menos uma verdadeiro positivo: {}".format(one_hit))
    print("Pelo menos um falso positivo: {}".format(one_mistake))
    print("True positives: {}".format(tp))
    print("False positives: {}".format(fp))
    print("True negatives: {}".format(tn))
    print("False negatives: {}".format(fn))
    print('Hamming loss:', hm)


    return {'One hit': one_hit, 'One mistake': one_mistake, 'TP': tp, 'FP': fp, 'TN':tn, 'FN':fn, 'HL': hm}



import pickle
import numpy as np
from .utils import (GpamTokenizer,
                   cnn_pecas_model2,
                   CallBack,
                   Y_transform,
                   binarize_pred
                   )
from sklearn.model_selection import train_test_split
from sklearn.metrics import classification_report
from sklearn import preprocessing

DEFAULT_VOCAB = pickle.loads(
    open("./default_vocab/vocab_112_bag.pk", "rb").read()
)
TYPE_PECAS = "Tag Mapeada"
BODY = "text"
MAX_FEATURES = len(DEFAULT_VOCAB)
SEQUENCE_LEN = 100
EMBEDDING_OUT = 100


class PecasModel:
    def __init__(self, dataframe, vocab=DEFAULT_VOCAB, classifier=None):
        self.dataframe = dataframe
        self.vocab = vocab
        if classifier is None:
            n_classes = len(self.dataframe[TYPE_PECAS].unique())
            self.classifier = cnn_pecas_model2(n_classes, MAX_FEATURES,
                                              SEQUENCE_LEN, EMBEDDING_OUT)

        else:
            self.classifier = classifier

    def _split(self, x, y):
        X_train, X_test, Y_train, Y_test = train_test_split(
            x,
            y,
            test_size=0.3,
            random_state=0)

        return X_train, X_test, Y_train, Y_test

    def _tokenize(self, X_train, shape=100):
        tokenizer = GpamTokenizer(self.vocab, X_train)
        return tokenizer.tokenizer_with_vocab(shape)

    def train(self, split_df=False, batch_size=500, epochs=20):
        print("Training...")
        X_train, Y_train = (self.dataframe[BODY], self.dataframe[TYPE_PECAS])

        if split_df:
            X_train, X_test, Y_train, Y_test = self._split(X_train, Y_train)
        y_transform = Y_transform(self.dataframe[TYPE_PECAS])
        Y_train = y_transform.transform(Y_train)
        vector = self._tokenize(X_train)
        callback = CallBack()
        self.classifier.fit(vector, Y_train, batch_size=batch_size,
                            epochs=epochs, callbacks=[callback])
        if split_df:
            return self.metrics(X_test, Y_test)

    def return_model(self):
        model = self.classifier.to_json()
        self.classifier.save_weights("/tmp/weights.h5")
        weights_file = open("/tmp/weights.h5", "rb")
        weights = weights_file.read()
        weights_file.close()
        return model, weights

    def metrics(self, X_test, Y_test):
        X_test = self._tokenize(X_test)
        pred = self.classifier.predict(X_test, verbose=1)
        new_pred = binarize_pred(pred)
        
        lb = preprocessing.LabelBinarizer()
        lb.fit(list(set(self.dataframe["Tag Mapeada"])))
        lb.inverse_transform(np.asarray(new_pred))
        
        le = preprocessing.LabelEncoder()
        le.fit(list(set(self.dataframe['Tag Mapeada'])))
        a = [each for each in enumerate(le.classes_)] 
        print(a)
        pred_y = le.transform(lb.inverse_transform(np.asarray(new_pred)))

        y_transform = Y_transform(self.dataframe[TYPE_PECAS])
        Y_test = y_transform.transform(Y_test)
        Y_test = le.transform(lb.inverse_transform(Y_test))

        return classification_report(Y_test, pred_y, output_dict=True)




from tqdm import tqdm
import numpy as np
import tensorflow as tf
from tensorflow import keras
from sklearn import preprocessing


class GpamTokenizer:
    def __init__(self, vocab, texts):
        self.vocab = vocab
        self.texts = texts
        self.dict_vocab = {}

    def list_2_dict(self):
        id = 1
        for word in self.vocab:
            self.dict_vocab[word] = id
            id += 1

    def return_tokens(self, text):
        return text.split(" ")

    def transform_tokens(self, tokens):
        result_transform = []
        for token in tokens:
            try:
                id = self.dict_vocab[token]
                result_transform.append(id)
            except Exception:
                continue

        return result_transform

    def pad_vector(self, texts, num):
        reshape_v = []

        for each in tqdm(texts):
            if len(each) >= num:
                reshape_v.append(each[0:num])
            else:
                zeros = num - len(each)
                temp = each
                v_zeros = [0 for each in range(zeros)]
                temp.extend(v_zeros)
                reshape_v.append(temp)

        return reshape_v

    def tokenizer_with_vocab(self, num):
        self.list_2_dict()
        result_texts = []
        for i in tqdm(range(len(self.texts))):
            tokens = self.return_tokens(self.texts.iloc[i])
            result_transform = self.transform_tokens(tokens)
            result_texts.append(result_transform)

        result = self.pad_vector(result_texts, num)

        return np.matrix(result)


def binarize_pred(preds):
    new_preds = []
    for each in preds:
        num = max(each)
        new_list = np.asarray([0 for temp in range(len(each))])
        new_list[np.where(each == num)[0][0]] = 1
        new_preds.append(np.asarray(new_list))
    return new_preds


def cnn_pecas_model(n_class, max_features, sequence_len, embedding_out):
    model = keras.Sequential()
    model.add(keras.layers.Embedding(
        input_dim=max_features,
        output_dim=embedding_out,
        input_length=sequence_len))

    model.add(keras.layers.Conv1D(kernel_size=4, filters=256, padding='same'))
    model.add(keras.layers.Dropout(0.5))
    model.add(keras.layers.MaxPooling1D(pool_size=2))
    model.add(keras.layers.Flatten())
    model.add(keras.layers.Dense(512, activation='relu'))
    model.add(keras.layers.Dropout(0.5))
    model.add(keras.layers.Dense(n_class, activation='softmax'))
    model.compile(loss='categorical_crossentropy', optimizer='adam',
                  metrics=['accuracy'])

    return model

def cnn_pecas_model2(n_class, max_features, sequence_len, embedding_out):
    f1_base = keras.layers.Input(shape=(sequence_len, ), dtype='int32')
    text_embedding = keras.layers.Embedding(input_dim=max_features, output_dim=embedding_out,
                               input_length=sequence_len)(f1_base)

    filter_sizes = [3, 4, 5]
    convs = []
    for filter_size in filter_sizes:
        l_conv = keras.layers.Conv1D(filters=256, kernel_size=filter_size, padding='same', activation='relu')(text_embedding)
        l_batch = keras.layers.BatchNormalization()(l_conv)
        l_pool = keras.layers.MaxPooling1D(2)(l_conv)

        convs.append(l_pool)

    l_merge = keras.layers.Concatenate(axis=1)(convs)
    l_pool1 = keras.layers.MaxPooling1D(50)(l_merge)
    l_flat = keras.layers.Flatten()(l_pool1)
    l_dense = keras.layers.Dense(128, activation='relu')(l_flat)
    x = keras.layers.Dropout(0.5)(l_dense)
    #f1_x = Flatten()(f1_x)
    x = keras.layers.Dense(n_class, activation='softmax')(x)
    model = keras.models.Model(inputs=f1_base, outputs=x)
    # determine Loss function and Optimizer
    model.compile(loss='categorical_crossentropy', optimizer='adam', metrics=['accuracy'])
    return model


class Y_transform:
    def __init__(self, y_all):
        self.y_all = y_all
        self.le = preprocessing.LabelEncoder()
        self.le.fit(list(set(y_all)))

    def transform(self, y):
        y_temp = self.le.transform(y)
        y_temp = np.transpose(y_temp)
        y_temp = tf.keras.utils.to_categorical(y_temp)
        return y_temp


class CallBack(tf.keras.callbacks.Callback):
    def on_epoch_end(self, epoch, logs={}):
        if(logs.get('acc') and logs.get('acc') >= 0.98):
            print("\nReached 95% accuracy so cancelling training!")
            self.model.stop_training = True
